import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import layers , forward_model , Losses , backward_model
from dataset import dataset
import math
from optimization import momentum ,ADAM
from Metric import evaluation_metrics
from matplotlib import style
import matplotlib; matplotlib.use("TkAgg")
style.use('fivethirtyeight')
fig = plt.figure()
ax1 = fig.add_subplot(1, 1, 1)
#from plotting import animate
class training_model:
    '''
    " The class concerned the training using mini_batch gradient descent with optimizations like adam and momnetum GD "
    '''
    def __init__(self,minibatches, activations,alpha=0.0007 , no_of_iterations=25, print_cost = True, lambd = 0 ,momentum=0,beta=0.9,ADAM=1):
        '''
        :param minibatches : list of mini_batches after dividing the training set to mini_batches
        :param activations : the type of activation function in each layer
        :param alpha: learning rate
        :param no_of_iterations: number of iterations
        :param print_cost: boolean variable , put it with ( True ) value if you want to print the cost every 10 iterations
        :param lambd: regularization parameter
        :param momentum : boolean : put it with ( True ) value if you want to apply the momentum Gradient descent
        :param beta : momentum parameter
        :param ADAM : boolean : put it with a ( True ) value if you want to apply ADAM optimization
        :return : The Trained Parameters for a certain model trained on a certain dataset
        '''
        self.input = minibatches
        self.activation_functions = activations
        self.Beta = beta
        self.learning_rate = alpha
        self.no_of_iterations = no_of_iterations
        self.print_cost = print_cost
        self.regularization_parameter = lambd
        self.momentum_or_no = momentum
        self.adam_or_not = ADAM

    def update_parameters(self, parameters, grads, learning_rate):
        '''
        "The function which is used to update the weights and biases of the model with Vanilla Gradient descent ."
        :param parameters: weights and biases of the past iteration ( before updating )
        :param grads: the Gradients of the weights and biases , the output of the backward propagation
        :param learning_rate: The learning rate
        :return: parameters : updated weights and biases after completing one iteration of the training
        '''
        L = len(parameters) // 2  # number of layers in the neural network
        for l in range(L):
            parameters["W" + str(l + 1)] = parameters["W" + str(l + 1)] - learning_rate * grads["dW" + str(l + 1)]
            parameters["b" + str(l + 1)] = parameters["b" + str(l + 1)] - learning_rate * grads["db" + str(l + 1)]

        return parameters

    def animate(i):
        '''
        " A function which is used to draw a live plotting of the Cost function during the training process "
        '''
        # print('hi')
        graph_data = open('costs.txt', 'r').read()
        lines = graph_data.split('\n')
        xs = []
        ys = []
        for line in lines:
            if len(line) > 1:
                x, y = line.split(',')
                xs.append(float(x))
                ys.append(float(y))
        ax1.clear()
        plt.ylabel('cost')
        plt.xlabel('iterations')
        ax1.plot(xs, ys)
    def train(self):
        '''
        This function considered as the integration of all the past Modules together to start training any model
        the deep learning engineer will have the option to choose :
        1- the activation function of each layer
        2- the loss type
        3- the number of iterations
        this function will plot live graph for the training cost and finally will print the the accuracy resulted from the test set training using the parameters resulted from the training set training .
        :return: The Trained parameters
        '''
        adam_flag=1
        layers_dimensions = [self.input[0][0].shape[0], 128, 10]
        temp_layers = layers.layers(layers_dimensions)
        parameters = temp_layers.layers_init()
        temp = parameters
        if (self.momentum_or_no):
            velocity=momentum(parameters).velocity_preparation()
        if(self.adam_or_not):
            exponentially_weighted_parameter,RMS_parameter= ADAM(parameters).adam_preparation()
        cost_file = open("costs.txt", 'a+')
        cost_file.truncate(0)
        cost_file.close()
        for i in range(0,self.no_of_iterations):
            for j in range(len(self.input)):
                train_Y = data_set.labels_to_onehot(self.input[i][1])
                train_X = self.input[i][0]
                no_of_training_examples=self.input[i][1].shape[1]
                predictions, packet_of_packets = forward_model.forward_model().forward_model(train_X, parameters,self.activation_functions)
                if self.regularization_parameter == 0:
                    cost = Losses.multiclass_loss(train_Y, predictions).cost()
                else:
                    cost = Losses.regularization().compute_cost_with_regularization(predictions, train_Y, parameters,self.regularization_parameter,"multiclass")


                grads = backward_model.model_backward_general(predictions, train_Y, packet_of_packets, "multiclass",self.regularization_parameter,self.activation_functions).model_backward()
                if(self.momentum_or_no):
                    parameters, velocity = momentum(parameters).update_with_momentum(velocity, self.learning_rate,
                                                                                     self.Beta, grads)

                elif(self.adam_or_not):
                    parameters,exponentially_weighted_parameter,RMS_parameter=ADAM(parameters).update_with_adam(exponentially_weighted_parameter,RMS_parameter,self.learning_rate,parameters,grads,i)

                else:
                    parameters = self.update_parameters(parameters, grads, self.learning_rate)
            cost_file = open("costs.txt", 'a+')
            cost_file.write(f"{i},{cost} \n")
            cost_file.close()
            plt.ion()
            plt.show()
            plt.draw()
            plt.pause(1)
            print(f"cost after epoch{i}: {cost}")

        return parameters

    def random_mini_batches(X, Y, mini_batch_size=2000):
        '''
        " this function will divide the whole training examples to small modules each one called ( mini_batch )
        :param X : The whole Training Set AKA all the training examples
        :param Y : The whole Training examples' labels
        :param mini_batch_size: The wanted mini_batch size for the training examples
        :return : a list of mini batches after division
        '''

        m = X.shape[1]  # number of training examples
        mini_batches = []

        permutation = list(np.random.permutation(m))
        shuffled_X = X[:, permutation]
        shuffled_Y = Y[:, permutation].reshape((1, m))

        num_complete_minibatches = math.floor(
            m / mini_batch_size)
        for k in range(0, num_complete_minibatches):
            mini_batch_X = shuffled_X[:, k * mini_batch_size: (k + 1) * mini_batch_size]
            mini_batch_Y = shuffled_Y[:, k * mini_batch_size: (k + 1) * mini_batch_size]
            mini_batch = (mini_batch_X, mini_batch_Y)
            mini_batches.append(mini_batch)

        if m % mini_batch_size != 0:
            mini_batch_X = shuffled_X[:, int(m / mini_batch_size) * mini_batch_size:]
            mini_batch_Y = shuffled_Y[:, int(m / mini_batch_size) * mini_batch_size:]
            mini_batch = (mini_batch_X, mini_batch_Y)
            mini_batches.append(mini_batch)

        return mini_batches





import numpy as np
import Losses
import activations
class backward_model:
    '''
    " The class which is concerned with calculating the back propagation gradients , which will be used to update the
        weights and biases through the training process
    '''
    def identity_backward(self,dz,linear_packet,lambd):
        '''
        :param dz: The derivative of the loss w.r.t the input to the activation function (Z)
        :param linear_packet: a tuple which contains ( the input , weights and biases of the current layer )
        :param lambd: the Regularization parameter
        :return: - delta_input_previous , the gradient of the past input
                 - delta_w : gradient of the weights of the current layer
                 - delta_b : the gradient of the biases of the current layer
        '''
        A_prev, W, b = linear_packet
        m = A_prev.shape[1]
        if lambd:
            delta_w = (1 / m) * np.dot(dz, A_prev.T) + (lambd / m) * W  # derivative of loss w.r.t (w)
        else:
            delta_w = (1 / m) * np.dot(dz, A_prev.T)  # derivative of loss w.r.t (w)
        #delta_w =np.dot(dz, A_prev.T)  # derivative of loss w.r.t (w)
        delta_b = (1 / m) * np.sum(dz, axis=1, keepdims=True) # derivative of loss w.r.t (b)
        delta_input_prev = np.dot(W.T, dz)  # derivative of loss w.r.t (input)
        return delta_input_prev, delta_w, delta_b


    def activation_backward(self,delta_A,packet_of_packets, activation_type,lambd):
        '''
        :param delta_A: the derivative of the loss function w.r.t the activation function
        :param packet_of_packets: Tuple of 2 elements which will be used in backward propagation :
                     1- linear packer : contains ( input , weights , biases ) of the current layer
                     2- activation packet : contains ( Z ) which is the input to the activation function
        :param activation_type: the type of the activation function used in this layer
        :param lambd: regularization parameter
        :return: - delta_input_previous , the gradient of the past input
                 - delta_w : gradient of the weights of the current layer
                 - delta_b : the gradient of the biases of the current layer
        '''
        linear_packet, act_packet = packet_of_packets

        if activation_type == "relu":
            #print("hi")
            temp = activations.relu()
            dZ = temp.backaward(delta_A, act_packet)    # we have to implement this relu backward function
            dA_prev, dW, db = self.identity_backward(dZ,linear_packet,lambd)
        elif activation_type == "sigmoid":
            #print("hi")
            temp = activations.Sigmoid()
            dZ = temp.backward(delta_A, act_packet)
            dA_prev, dW, db = self.identity_backward(dZ,linear_packet,lambd)
        # we will start from here tomorrow , we have to deal with Y_hat , y_true while creating instance from cost class
       # temp = Losses.square_difference()
        #dA = temp.backprop_cost(self.linear_packet)
        elif activation_type == "softmax":
            temp = activations.Softmax()
            dZ = temp.diff(delta_A)
            dA_prev, dW, db = self.identity_backward(dZ, linear_packet,lambd)

        return dA_prev, dW, db



'''
dA = np.array([[0.208,0.613]]).T
z = np.array([[0.702,0.867]]).T
A = np.array([[0.808,0.604]]).T
W = np.array([[0.323,0.076],[0.614,0.106]]).T
b = np.array([[0.676,0.360]]).T
linear_packet = (A,W,b)
act_packet = z
packet_of_packets = linear_packet , act_packet
da_prev , dW , dB = backward_model().activation_backward(dA,packet_of_packets,"sigmoid")
print(f"da_prev = {da_prev} ")
print(f"dW = {dW} ")
print(f"db = {dB} ")
'''


import numpy as np
class momentum:
    '''
    " The class concerned The momentum optimization "
    '''

    def __init__(self,parameters):
        '''
        :param parameters: the Weights and biases of The current model
        '''
        self.parameters=parameters


    def velocity_preparation(self):
        '''
        :return: Velocity parameter for weights and biases as to update the parameters using momentum GD
        '''
        weights = len(self.parameters) // 2  # number of layers in the neural networks
        v = {}
        # Initialize velocity
        for w in range(weights):
            v["dW" + str(w + 1)] = np.zeros((self.parameters["W" + str(w + 1)].shape[0], self.parameters["W" + str(w + 1)].shape[1]))
            v["db" + str(w + 1)] = np.zeros((self.parameters["b" + str(w + 1)].shape[0], self.parameters["b" + str(w + 1)].shape[1]))
        return v


    def update_with_momentum(self,velocity,learning_rate,exponentially_weighted_average_parameter,gradients):
        '''
        :param velocity: Velocity of the gradient with momentum
        :param learning_rate: the learning rate
        :param exponentially_weighted_average_parameter: Beta , the momentum hyperparameter
        :param gradients: dW , dB , the gradients of the weights and biases of each layer
        :return: parameters: updated weights and biases
                 velocity : updated velocity to be used in the next iteration
        '''
        L = len(self.parameters) // 2  # number of layers in the neural networks
        for l in range(L):
            # compute velocities
             velocity["dW" + str(l + 1)] = exponentially_weighted_average_parameter * velocity["dW" + str(l + 1)] + (1 - exponentially_weighted_average_parameter) *  gradients['dW' + str(l + 1)]
             velocity["db" + str(l + 1)] = exponentially_weighted_average_parameter *  velocity["db" + str(l + 1)] + (1 - exponentially_weighted_average_parameter) * gradients['db' + str(l + 1)]
            # update parameters
             self.parameters["W" + str(l + 1)] = self.parameters["W" + str(l + 1)] -  learning_rate *  velocity["dW" + str(l + 1)]
             self.parameters["b" + str(l + 1)] = self.parameters["b" + str(l + 1)] -  learning_rate *  velocity["db" + str(l + 1)]

        return self.parameters, velocity




class ADAM:
    '''
    " The class concerned ADAM optimization technique "
    '''
    def __init__(self,parameters):
        '''
        :param parameters: The weights and biases for each layer before any updates
        '''
        self.parameters=parameters


    def adam_preparation(self):
        '''
        " Initialization of ADAM optimization's parameters"
        :return: EWA : exponentially weighted average parameter
                 RMS : Root mean square prop parameter
        '''
        L = len(self.parameters) // 2  # number of layers in the neural networks
        EWA = {}
        RMS = {}

        # Initialize v, s. Input: "parameters". Outputs: "v, s".
        for l in range(L):
            # exponentially weighted average parameters
            EWA["dW" + str(l + 1)] = np.zeros((self.parameters["W" + str(l + 1)].shape[0], self.parameters["W" + str(l + 1)].shape[1]))
            EWA["db" + str(l + 1)] = np.zeros((self.parameters["b" + str(l + 1)].shape[0], self.parameters["b" + str(l + 1)].shape[1]))
            # RMS prop average parameters
            RMS["dW" + str(l + 1)] = np.zeros((self.parameters["W" + str(l + 1)].shape[0], self.parameters["W" + str(l + 1)].shape[1]))
            RMS["db" + str(l + 1)] = np.zeros((self.parameters["b" + str(l + 1)].shape[0], self.parameters["b" + str(l + 1)].shape[1]))

        return EWA, RMS

    def update_with_adam(self,EWA,RMS,learning_rate,parameters,gradients,epoch_num,frist_beta=0.9,second_beta=0.995,epsilon=1e-8,):
        '''
        :param EWA: Exponentially weighted average parameter
        :param RMS: RMS prop parameter
        :param learning_rate: the learning rate
        :param parameters: weights and biases before update
        :param gradients: the Gradients of weights and biases for the current model's layers
        :param epoch_num: the Epoch number
        :param frist_beta: Beta_1 , the first hyperparameter for Exponentially weighted average's parameter
        :param second_beta:Beta_2 , the 2nd hyperparameter for RMS prop's parameter
        :param epsilon: The safety margin as to not divide by zero
        :return: :param parameters : The updated parameters of the current model
                 :param: EWA : the updated Exponentially weighted average parameters
                 :param RMS : the updated RMS prop parameters
        '''

        L = len(parameters) // 2  # number of layers in the neural networks
        EWA_corrected = {}  # Initializing first moment estimate, python dictionary
        RMS_corrected = {}  # Initializing second moment estimate, python dictionary

        # Perform Adam update on all parameters
        for l in range(L):

            EWA["dW" + str(l + 1)] = frist_beta * EWA["dW" + str(l + 1)] + (1 - frist_beta) * gradients['dW' + str(l + 1)]
            EWA["db" + str(l + 1)] = frist_beta * EWA["db" + str(l + 1)] + (1 - frist_beta) * gradients['db' + str(l + 1)]
            #EWA_corrected["dW" + str(l + 1)] = EWA["dW" + str(l + 1)] / ((1 - frist_beta ** epoch_num)+epsilon)
            #EWA_corrected["db" + str(l + 1)] = EWA["db" + str(l + 1)] / ((1 - frist_beta ** epoch_num)+epsilon)
            RMS["dW" + str(l + 1)] = second_beta * RMS["dW" + str(l + 1)] + (1 - second_beta) * np.square(gradients['dW' + str(l + 1)])
            RMS["db" + str(l + 1)] = second_beta * RMS["db" + str(l + 1)] + (1 - second_beta) * np.square(gradients['db' + str(l + 1)])
            #RMS_corrected["dW" + str(l + 1)] = RMS["dW" + str(l + 1)] / ((1 - second_beta ** epoch_num)+epsilon)
            #RMS_corrected["db" + str(l + 1)] = RMS["db" + str(l + 1)] / ((1 - second_beta ** epoch_num)+epsilon)
            parameters["W" + str(l + 1)] = parameters["W" + str(l + 1)] - learning_rate * EWA["dW" + str(l + 1)] / (np.sqrt(RMS["dW" + str(l + 1)]) + epsilon)
            parameters["b" + str(l + 1)] = parameters["b" + str(l + 1)] - learning_rate * EWA["db" + str(l + 1)] / (np.sqrt(RMS["db" + str(l + 1)]) + epsilon)


        return parameters, EWA,RMS



import numpy as np
class square_difference():
    '''
    "This is the class which is Concerned with Calculating the SSD loss & it's derivative "
    '''
    def __init__(self,true_label,Y_hat):
        '''
        :param true_label: The true classes of the input training examples
        :param Y_hat: the output of the forward propagation .. the prediction of Training .. the output from the
                        activation function
        '''
        self.true_label=true_label
        self.Y_hat=Y_hat
    def cost(self):
        '''
        :return: The value of the cost using the SSD Loss function
        '''
        #cost =0.5 * np.mean(np.dot(np.transpose((self.true_label - self.Y_hat)), (self.true_label - self.Y_hat)) * (1/self.true_label.shape[0]))
        m = self.true_label.shape[1]
        #print(self.Y_hat.shape)
        logprobs = np.multiply(-np.log(self.Y_hat), self.true_label) + np.multiply(-np.log(1 - self.Y_hat), 1 - self.true_label)
        cost = 1. / m * np.nansum(logprobs)
        #cost = np.nansum(logprobs)
        return cost
    def backprop_cost(self,cache):
        '''
        :param cache: a tuple which contains the values of ( input , weight , bias ) of the current layer
        :return: returns the back propagation loss value for back propagation calculations
        '''
        X , W , b = cache
        #def_cost = np.mean(np.dot(X,(self.Y_hat-self.true_label).T))
        def_cost = self.Y_hat - self.true_label
        return def_cost     # dA

class regularization:
    '''
    "Implementing the cost function with L2 regularization."
    '''
    def compute_cost_with_regularization(self,prediction, Y, parameters, lambd,loss_type):
        '''
        :param prediction  -- post-activation, output of forward propagation, of shape (output size, number of examples)
        :param Y -- "true" labels vector, of shape (output size, number of examples)
        :param parameters -- python dictionary containing parameters of the model
        :return: cost - value of the regularized loss function (formula (2))
        '''
        m = Y.shape[1]
        x=0
        temp = len(parameters) //2
        if (loss_type=="SSD"):
            #print("hi")
            cross_entropy_cost = square_difference(Y,prediction).cost()  # This gives you the cross-entropy part of the cost
        elif (loss_type=="multiclass"):
                # print("hi")
                cross_entropy_cost = multiclass_loss(Y,prediction).cost()  # This gives you the cross-entropy part of the cost
        for i in range(1,temp):
           x=+np.sum(np.square (parameters["W"+str(i)]))
        ### START CODE HERE ### (approx. 1 line)
        L2_regularization_cost = lambd / (2 * m) * x
        ### END CODER HERE ###

        cost = cross_entropy_cost + L2_regularization_cost

        return cost
class loglikelihood():
    '''
    "This is the class which is Concerned with Calculating the loglikelihood loss & it's derivative "
    '''
    def __init__(self, true_label, Y_hat):
        '''
       :param true_label: The true classes of the input training examples
       :param Y_hat: the output of the forward propagation .. the prediction of Training .. the output from the
                        activation function
        '''
        self.true_label = true_label
        self.Y_hat = Y_hat
    def cost(self):
        '''
         :return: The value of the cost using the loglikelihood Loss function
        '''
        cost = 0.5*(np.mean(-1*np.log(np.abs((self.true_label/2)-0.5+self.Y_hat))))
        return cost
    def backprop_cost(self,cache):
        '''
        :param cache: a tuple which contains the values of ( input , weight , bias ) of the current layer
        :return: returns the back propagation loss value for back propagation calculations
        '''
        X , W , b = cache
        def_cost= np.mean((np.dot(-1*self.Y_hat*X)/(1+np.exp(np.dot(self.true_label,self.Y_hat)))))
        return def_cost



class multiclass_loss():
    '''
    "This is the class which is Concerned with Calculating the multiclass loss & it's derivative "
    '''
    def __init__(self, Y, Y_hat):
        '''
        :param true_label: The true classes of the input training examples
        :param Y_hat: the output of the forward propagation .. the prediction of Training .. the output from the
                              activation function
        '''
        self.Y = Y
        self.Y_hat = Y_hat
    def cost(self):
        '''
        :return: The value of the cost using the multiclass Loss function
        '''
        #m = self.Y.shape[1]
       # logprobs = np.multiply(-np.log(self.Y_hat), self.Y) + np.multiply(-np.log(1 - self.Y_hat), 1 - self.Y)
        #cost = 1. / m * np.nansum(logprobs)
        cost = -(1/ self.Y.shape[1]) * np.sum(np.multiply(self.Y, np.log(self.Y_hat)))
        return cost
    
    def backprop_cost(self,cache):
        '''
        :param cache: a tuple which contains the values of ( input , weight , bias ) of the current layer
        :return: returns the back propagation loss value for back propagation calculations
        '''
        X , W , b = cache
        #def_cost = np.mean(np.dot(X,(self.Y_hat-self.true_label).T))
        def_cost = self.Y_hat - self.Y
        return def_cost  



class identity_loglikelihood():
    '''
    "The class which is concerned with calculating the loglikelihood loss function with identity activation function"
    '''
    def __init__(self, true_label, Y_hat):
        '''
        :param true_label: The true classes of the input training examples
        :param Y_hat: the output of the forward propagation .. the prediction of Training .. the output from the
                                      activation function
        '''
        self.true_label = true_label
        self.Y_hat = Y_hat
    def cost(self):
        '''
        :return: the Cost value of the loglikelihood loss with an identity activation
        '''
        cost =  (np.mean(np.log(1+(np.exp(-1*self.true_label*self.Y_hat)))))
        return cost


class perceptron_model_loss():
    '''
    "The perceptron criteria loss Class"
    '''
    def __init__(self, true_label, Y_hat):
        '''
        :param true_label: The true classes of the input training examples
        :param Y_hat: the output of the forward propagation .. the prediction of Training .. the output from the
                                              activation function
        '''
        self.true_label = true_label
        self.Y_hat = Y_hat
    def cost(self):
        '''
        :return: the Cost value of the perceptron model loss with an identity activation
        '''
        cost = np.argmax(0,-1*np.dot(self.true_label,self.Y_hat))
        return cost


class hinge_loss():
    '''
    "The hinge_loss (SVM) loss Class"
    '''
    def __init__(self, true_label, Y_hat):
        '''
        :param true_label: The true classes of the input training examples
        :param Y_hat: the output of the forward propagation .. the prediction of Training .. the output from the
                                                      activation function
        '''
        self.true_label = true_label
        self.Y_hat = Y_hat
    def cost(self):
        '''
        :return : the Cost value of the loss with an identity activation
        '''
        cost = np.argmax(0,(-1*np.dot(self.true_label,self.Y_hat))+1)
        return cost

'''
y_hat=np.array([0.5,0.35,0.83,0.73])
#y_hat=y_hat.T
labels=np.array([1,1,1,-1])
test_loss=loglikelihood(labels,y_hat)
print(test_loss.cost())
'''
'''
y_hat=np.array([0,-0.6,1.6,1])
#y_hat=y_hat.T
labels=np.array([1,1,1,-1])
test_loss=identity_loglikelihood(labels,y_hat)
print(test_loss.cost())
'''


import numpy as np
import matplotlib.pyplot as plt
import Losses
import activations
import back_prop
import dataset
import forward_model
import backward_model
import forward_prop
import back_prop
import optimization
import Metric
import matplotlib
import layers
import init
import Batch
import Mini_batch
import matplotlib.animation as animation
from dataset import dataset
from Metric import evaluation_metrics
import matplotlib; matplotlib.use("TkAgg")
from optimization import momentum ,ADAM
from matplotlib import style


import numpy as np
import back_prop
import Losses
class model_backward_general:
    '''
    " The Class which is concerned with combining the back propagation functions to form an integrated backprop model ."
    '''
    def __init__(self,Predictions,true_label,packet_of_packets,loss_function,lambd,activation_functions):
        '''
        :param Predictions: Y-hat --> the output of the activation function
        :param true_label: the true classes' values (labels)
        :param activation_functions : the type of activation function in each layer
        :param packet_of_packets: Tuple of 2 elements which will be used in backward propagation :
                     1- linear packer : contains ( input , weights , biases ) of the current layer
                     2- activation packet : contains ( Z ) which is the input to the activation function
        :param loss_function: type of the loss function
        :param lambd: regularization parameter
        :return : gradients of weights and biases
        '''
        self.lambd=lambd
        self.activation_functions = activation_functions
        self.predictions = Predictions
        self.Y = true_label
        self.packet_of_packets=packet_of_packets
        self.loss_function_type=loss_function
    def model_backward(self):
        gradients = {}
        L = len(self.packet_of_packets)  # the number of layers
        # m = self.predictions.shape[1]
        #  Y = self.Y.reshape(self.predictions.shape)  # after this line, Y is the same shape as AL

        # Initializing the backpropagation by getting the derivative of the cost function we are using wrt the output ( dl/dA )

        #  dAL = - (np.divide(Y, AL) - np.divide(1 - Y, 1 - AL))  # derivative of cost with respect to AL
        #   dAL = np.mean(np.dot(X.T,(self.predictions-self.Y)))
        temp = back_prop.backward_model()
        current_cache = self.packet_of_packets[L - 1]

        # if (self.loss_function_type=="SSD"):
        #     temp_loss = Losses.square_difference(self.Y, self.predictions)
        #     dAL = temp_loss.backprop_cost(current_cache[0])
        # elif (self.loss_fuction_type=="LLH"):
        #     temp_loss = Losses.loglikelihood(self.Y, self.predictions)
        #     dAL = temp_loss.backprop_cost(current_cache[0])
        if (self.loss_function_type=="SSD"):
            temp_loss = Losses.square_difference(self.Y, self.predictions)
            dAL = temp_loss.backprop_cost(current_cache[0])
        elif (self.loss_function_type=="multiclass"):
            temp_loss = Losses.multiclass_loss(self.Y, self.predictions)
            dAL = temp_loss.backprop_cost(current_cache[0])

        gradients["dA" + str(L - 1)], gradients["dW" + str(L)], gradients["db" + str(L)] = temp.activation_backward(dAL,current_cache,"softmax",self.lambd)

        # Loop from l=L-2 to l=0
        for l in reversed(range(L - 1)):
            # lth layer: (RELU -> LINEAR) gradients.
            # Inputs: "grads["dA" + str(l + 1)], current_cache". Outputs: "grads["dA" + str(l)] , grads["dW" + str(l + 1)] , grads["db" + str(l + 1)]
            current_cache = self.packet_of_packets[l]
            dA_prev_of_l, dW_of_l, db_of_l = temp.activation_backward(gradients["dA" + str(l + 1)], current_cache, self.activation_functions[l-1],self.lambd)
            gradients["dA" + str(l)] = dA_prev_of_l
            gradients["dW" + str(l + 1)] = dW_of_l
            gradients["db" + str(l + 1)] = db_of_l

        return gradients

    '''
    def model_backward_with_droupout(self):
        gradients = {}
        #        linear_packet , activation_packet = self.packet_of_packets
        L = len(self.packet_of_packets)  # the number of layers
        # m = self.predictions.shape[1]
        #  Y = self.Y.reshape(self.predictions.shape)  # after this line, Y is the same shape as AL

        # Initializing the backpropagation by getting the derivative of the cost function we are using wrt the output ( dl/dA )

        #  dAL = - (np.divide(Y, AL) - np.divide(1 - Y, 1 - AL))  # derivative of cost with respect to AL
        #   dAL = np.mean(np.dot(X.T,(self.predictions-self.Y)))
        # Lth layer (SIGMOID -> LINEAR) gradients. Inputs: "dAL, current_cache". Outputs: "grads["dAL-1"], grads["dWL"], grads["dbL"]
        # grads["dA" + str(L-1)], grads["dW" + str(L)], grads["db" + str(L)] = None
        temp = back_prop.backward_model()
        current_cache = self.packet_of_packets[L - 1]

        # if (self.loss_function_type=="SSD"):
        #     temp_loss = Losses.square_difference(self.Y, self.predictions)
        #     dAL = temp_loss.backprop_cost(current_cache[0])
        # elif (self.loss_fuction_type=="LLH"):
        #     temp_loss = Losses.loglikelihood(self.Y, self.predictions)
        #     dAL = temp_loss.backprop_cost(current_cache[0])
        if (self.loss_function_type == "SSD"):
            temp_loss = Losses.square_difference(self.Y, self.predictions)
            dAL = temp_loss.backprop_cost(current_cache[0])
        elif (self.loss_function_type == "multiclass"):
            temp_loss = Losses.multiclass_loss(self.Y, self.predictions)
            dAL = temp_loss.backprop_cost(current_cache[0])

        gradients["dA" + str(L - 1)], gradients["dW" + str(L)], gradients["db" + str(L)] = temp.activation_backward(dAL,current_cache,"softmax",self.lambd)

        # Loop from l=L-2 to l=0
        for l in reversed(range(L - 1)):
            # lth layer: (RELU -> LINEAR) gradients.
            # Inputs: "grads["dA" + str(l + 1)], current_cache". Outputs: "grads["dA" + str(l)] , grads["dW" + str(l + 1)] , grads["db" + str(l + 1)]
            current_cache = self.packet_of_packets[l]
            dA_prev_of_l, dW_of_l, db_of_l = temp.activation_backward(gradients["dA" + str(l + 1)], current_cache,"sigmoid", self.lambd)
            gradients["dA" + str(l)] = dA_prev_of_l
            gradients["dW" + str(l + 1)] = dW_of_l
            gradients["db" + str(l + 1)] = db_of_l

        return gradients
    '''



import numpy as np
import pandas as pd
from forward_model import forward_model

class evaluation_metrics:
    def __init__(self, labels, input,parameters,activation_functions):
        self.input = input
        self.Y_true = labels
        self.parameters=parameters
        self.activations=activation_functions

    def confusionMatrix(self):
        y_hat=[]
        predictions,packet_of_packets=forward_model().forward_model(self.input,self.parameters,self.activations)
        predictions=predictions.T
        for i in range(predictions.shape[0]):
            max=np.argmax(predictions[i])
            y_hat.append(max)

        classes = set(self.Y_true[0])
        number_of_classes = len(classes)
        conf_matrix = pd.DataFrame(
            np.zeros((number_of_classes, number_of_classes), dtype=int),
            index=classes,
            columns=classes)
        for i, j in zip(self.Y_true[0], y_hat):
            conf_matrix.loc[i, j] += 1
        return conf_matrix.values, conf_matrix

    def TP(self):
        values, cm = self.confusionMatrix()
        return np.diag(cm)

    def FP(self):
        values, cm = self.confusionMatrix()
        return np.sum(cm, axis=0) - self.TP()

    def FN(self):
        values, cm = self.confusionMatrix()
        return np.sum(cm, axis=1) - self.TP()

    def Accuracy(self, data_size):
        return np.sum(self.TP()/data_size)

    def Precision(self):
        return np.mean(self.TP() / (self.TP() + self.FP()))

    def Recall(self):
        return np.mean(self.TP() / (self.TP() + self.FN()))

    def F1_score(self):
        if self.TP() > 0:
            return 2 * ((self.Precision() * self.Recall()) / (self.Precision() + self.Recall()))
        else:
            return 0






from layers import layers
from forward_prop import forward_prop
class forward_model():
    '''
    " Class of Applying Forward propagation process . "
    '''
    def forward_model (self,X,parameters,activations):
        '''
        :param X: The input to the current layer as to start the forward propagation process
        :param parameters: Weights and biases of the current layer
        :param activations: list of the activation functions of each layer
        :return:  - Y_hat ( predictions resulted from the current layer's forward propagation )
                  - packet of packets : Tuple of 2 elements which will be used in backward propagation :
                     1- linear packer : contains ( input , weights , biases ) of the current layer
                     2- activation packet : contains ( Z ) which is the input to the activation function
        '''
        packet_of_packets = []
        A = X # as for the first time A will equal to the input layer
        L = len(parameters) // 2  # number of layers in the neural network
        test_forward=forward_prop()
        for l in range(1, L):
            A_prev = A
            A, packet = test_forward.activation_forward(A_prev, parameters['W' + str(l)], parameters['b' + str(l)], activation_type=activations[l])
            packet_of_packets.append(packet)

        prediction, packet = test_forward.activation_forward(A, parameters['W' + str(L)], parameters['b' + str(L)], activation_type="softmax")
        packet_of_packets.append(packet)



        return prediction, packet_of_packets





import pandas as pd
import numpy as np
import pickle
import os
class dataset:
    def __init__(self, dataset_name, path):
        self.dataset_name = dataset_name
        self.path = path

    def get_dataset(self):
        if self.dataset_name == 'mnist':
            training_set = pd.read_csv(self.path+'/train.csv/mnist_train.csv')
            testing_set = pd.read_csv(self.path+'/test.csv/mnist_test.csv')
            training_set = np.array(training_set)
            testing_set = np.array(testing_set)
            training_labels = np.array(training_set[:, 0]) # all rows of first columns
            training_examples = np.array(training_set[:, 1:]) #all data except first column
            training_labels = np.atleast_2d(training_labels)
            testing_labels = np.array(testing_set[:, 0])
            testing_examples = np.array(testing_set[:, 1:])
            testing_labels = np.atleast_2d(testing_labels)

            return training_examples/255, training_labels,  testing_examples/255, testing_labels # X , Y
        elif self.dataset_name == 'cifar_10':
            training_examples = []
            training_labels = []
            names = os.listdir(self.path)
            for name in names:
                if '_batch' in name:
                    with open(self.path+'/'+name, 'rb') as f:
                        data = pickle.load(f, encoding='bytes')
                        if 'test' in name: #test_batch
                            testing_examples = data[b'data']
                            testing_labels = data[b'labels']
                        else:
                            training_examples.append(data[b'data'])
                            training_labels.append(data[b'labels'])
            training_examples = np.array(training_examples)
            training_labels = np.array(testing_labels)
            training_labels = training_labels.reshape((-1, 1))
            training_examples = training_examples.reshape((-1,3072))
            testing_labels = np.atleast_2d(testing_labels)
            training_labels = np.atleast_2d(training_labels.T) # onerow
            return training_examples/255, training_labels,  testing_examples/255, testing_labels


    def labels_to_onehot(self,labels):
        max = np.max(labels)
        enc = np.zeros((max + 1, labels.shape[1]))
        for i in range(labels.shape[1]):
            for j in range(max):
                enc[j][i] = (labels[0][i] == j)  ###hot encoding trainLabels
                if (labels[0][i] == 9):
                    enc[9][i] = 1
        onehot_labels = enc
        return onehot_labels












































'''files = [
    ['training set images', 'train-images-idx3-ubyte.gz'],
    ['test set images', 't10k-images-idx3-ubyte.gz'],
    ['training set labels', 'train-labels-idx1-ubyte.gz'],
    ['test set labels', 't10k-labels-idx1-ubyte.gz']
]


def download():
    url = 'http://yann.lecun.com/exdb/mnist/'
    for file in files:
        urllib.urlretrieve(url + str(file[1]), file[0])


def save():
    data = {}
    for file in files[:2]:
        with gzip.open(file[1], 'rb') as f:
            data[file[0]] = np.frombuffer(f.read(), np.uint8, count=-1, offset=16).reshape(-1, 784)

    for file in files[-2:]:
        with gzip.open(file[1], 'rb') as f:
            data[file[0]] = np.frombuffer(f.read(), np.uint8, count=-1, offset=8)

    with open('mnst.pkl', 'wb') as f:
        pickle.dump(data, f)


def load():
    with open('mnst.pkl', 'wb') as f:
        data = pickle.load(f, encoding="UTF-8")
    return data


def batch(num_of_batches):
    data = load()
    X = data['training set images']
    Y = data['training set labels']
    X = X.reshape(num_of_batches, (len(X)/num_of_batches), len(X[0]))
    Y = Y.reshape(num_of_batches, (len(X)/num_of_batches), len(X[0]))


save()
load()
'''



































'''files = [
    ['training set images', 'train-images-idx3-ubyte.gz'],
    ['test set images', 't10k-images-idx3-ubyte.gz'],
    ['training set labels', 'train-labels-idx1-ubyte.gz'],
    ['test set labels', 't10k-labels-idx1-ubyte.gz']
]


def download():
    url = 'http://yann.lecun.com/exdb/mnist/'
    for file in files:
        urllib.urlretrieve(url + str(file[1]), file[0])


def save():
    data = {}
    for file in files[:2]:
        with gzip.open(file[1], 'rb') as f:
            data[file[0]] = np.frombuffer(f.read(), np.uint8, count=-1, offset=16).reshape(-1, 784)

    for file in files[-2:]:
        with gzip.open(file[1], 'rb') as f:
            data[file[0]] = np.frombuffer(f.read(), np.uint8, count=-1, offset=8)

    with open('mnst.pkl', 'wb') as f:
        pickle.dump(data, f)


def load():
    with open('mnst.pkl', 'wb') as f:
        data = pickle.load(f, encoding="UTF-8")
    return data


def batch(num_of_batches):
    data = load()
    X = data['training set images']
    Y = data['training set labels']
    X = X.reshape(num_of_batches, (len(X)/num_of_batches), len(X[0]))
    Y = Y.reshape(num_of_batches, (len(X)/num_of_batches), len(X[0]))


save()
load()
'''


import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib import style

style.use('fivethirtyeight')

fig = plt.figure()
ax1 = fig.add_subplot(1,1,1)

def animate(i):
    graph_data = open('our_plotting.txt','r').read()
    lines = graph_data.split('\n')
    xs = []
    ys = []
    for line in lines:
        if len(line) > 1:
            x, y = line.split(',')
            xs.append(float(x))
            ys.append(float(y))
    ax1.clear()
    ax1.plot(xs, ys)

ani = animation.FuncAnimation(fig, animate, interval=1000)
plt.show()



import numpy as np

import matplotlib.pyplot as plt

class Sigmoid:
    '''
    "The class concerned the Sigmoid function which will calculate it and it's derivative for the sake of forward
                and backward propagation respectively . "
    '''
    def forward(self, z):
        '''

        :param z: The result of multiplying ( Weight * input ) + bias
        :return: The sigmoid of the input
        '''
        sig = 1 / (1 + np.exp(-z))
        act_packet = z
        return sig, act_packet

    def derivative(self,z):
        '''
        :param z: The sigmoid input
        :return: The shortcut result of the derivative of the sigmoid function
        '''
        sig = 1 / (1 + np.exp(-z))
        return sig*(1-sig)
    def backward(self,dA,act_packet):
        '''

        :param dA: Derivative of the Loss function with respect the sigmoid function
        :param act_packet: The value of the Z , which is the output before entering the activation function
        :return:
        '''
        #derivative_of_sigmoid = self.derivative(act_packet)
        #dz=np.multiply(dA,derivative_of_sigmoid)
        #test = np.ones_like(dz)
        return dA




# backward activation here #

class relu:
    '''
    The class concerned the relu function which will calculate it and it's derivative for the sake of forward
                and backward propagation respectively .
    '''
    def forward(self, z):
        '''

        :param z: The relu input
        :return: the relu output for the input
        '''
        rel = np.maximum(0, z)
        act_packet = z
        return rel, act_packet
# backward action here
    def backaward(self,dLA, activation_packet):
        '''

        :param dLA: Derivative of the Loss function with respect the relu function
        :param activation_packet: The value of the Z , which is the output before entering the activation function
        :return: the derivative of the cost function with respect to Z
        '''

        Z = activation_packet
        #print("Z")
        #print(Z.shape)
        dZ = np.array(dLA, copy=True)  # just converting dz to a correct object.

        # When z <= 0, you should set dz to 0 as well.
        #dZ[Z<=0] = 0
        dZ = np.where(Z>0,1,0)
        assert (dZ.shape == Z.shape)
        return dZ
      #  print(type(Z))
        #dZ = np.array(dLA, copy=True)  # just converting dz to a correct object.
     #   print(type(dZ))
        # dl/dz = dl/da * da/dz .. as da/dz = 1 if the z > 0 when we are using relu activation function when z>0 , so dl/dz = dl/da if z>0
        # but if z<=0 so da/dz will equal 0 then dl/dz = 0
        # When z <= 0, you should set dz to 0 as well.


class leaky_relu:
    '''
    The class concerned the leaky_relu function which will calculate it and it's derivative for the sake of forward
                and backward propagation respectively .
    '''
    def forward(self,z):
        '''

        :param z: the leaky_relu input
        :return: the leaky_relu output of the input
        '''
        y1 = ((z > 0) * z)
        y2 = ((z <= 0) * z * 0.01)
        leaky= y1 + y2
        act_packet = z
        return leaky , act_packet

    def backaward(self, dLA, activation_packet):
        '''
        :param dLA: Derivative of the Loss function with respect the leaky_relu function
        :param activation_packet: The value of the Z , which is the output before entering the activation function
        :return: the derivative of the cost function with respect to Z
        '''

        Z = activation_packet
        # print("Z")
        # print(Z.shape)
        dZ = np.array(dLA, copy=True)  # just converting dz to a correct object.

        # When z <= 0, you should set dz to 0 as well.
        dZ[Z <= 0] = 0
        # dZ = np.where(Z>0,1,0)
        assert (dZ.shape == Z.shape)
        return dZ

class tanh:
    '''
     The class concerned the tanh function which will calculate it and it's derivative for the sake of forward
                and backward propagation respectively .
    '''
    def forward(self,z):
        '''
        :param z: the tanh input
        :return: the tanh output of the input
        '''
        TANH = np.tanh(z)
        act_packet = z
        return TANH , act_packet
    def diff(self,z):
        '''
        :param z : the input to the tanh function
        :return : the backward shortcut result of the derivative of the tanh function w.r.t the Z
        '''
        tanh=self.forward(z)
        return (1-np.square(tanh))
    def backward(self,dA,act_packet):
        '''

        :param dA: Derivative of the Loss function with respect the tanh function
        :param act_packet: The value of the Z , which is the output before entering the activation function
        :return: the derivative of the cost w.r.t  Z ( dL/dZ = dL/dA * dA/dZ )
        '''
        derivative_of_tanh = self.diff(act_packet)
        dz=np.multiply(dA,derivative_of_tanh)
        #test = np.ones_like(dz)
        return dz

class Softmax():
    '''
    '''
    def forward(self, z):
        '''
        :param z: the input to the softmax function
        :return: the forward output of the softmax activation function
        '''
        act_packet = z
        ans = np.exp(z - np.max(z))
        ans = ans / ans.sum(axis=0)
        return ans, act_packet


    def diff(self,dA):
        '''
        :param dA: derivative of Loss w.r.t the output of the softmax activation function
        :return: the derivative of the loss w.r.t Z --> the output before entering the activation function
        '''
        #dA = np.ones(z.shape)
        #m, n = z.shape
        #p = self.__call__(z)
        #tensor1 = np.einsum('ij,ik->ijk', p, p)  # (m, n, n)
        #tensor2 = np.einsum('ij,jk->ijk', p, np.eye(n, n))  # (m, n, n)
        #dz = tensor2 - tensor1
        #dz = np.einsum('ijk,ik->ij', dz, dA)
        #return dz
        return dA

'''
dummy=(np.arange(100)/10)-5
test_sigmoid = tanh()
result , act = test_sigmoid.forward(dummy)
#print(result)
plt.plot(dummy,result)
plt.show()
'''
'''
dummy=[3.1, -9.3, 7, 8.7, 3.6, 5.2, 4.7, -2.2, 3.1, -6.6]
test_soft_max=softmax()
result , act_packet = test_soft_max.forward(dummy)
print(result)
'''


import numpy as np

class initialization:
    '''
    " This is the initialization class which will initialize the weights and biases for the layers of any model . "
    '''
    def __init__(self,dimension_layers):
        self.dimension_layers = dimension_layers
        '''
        " The Constructor of this class takes the a list contains the number of hidden units in each layer ."  
        :type dimension_layers: List . 
        :param dimension_layers: list of the number of hidden units in each hidden layer .
        '''


    def initialize_parameters(self): # [6,5,9,1]

        '''
         " This function will initialize a random values for weights and zeros for biases using the
                            dimension_layers list . "

         :type parameters: dictionary .
         :param parameters: this dictionary contains the weights and biases with the same dimension as
                                each layers' dimension .
        '''
        parameters = {}
        np.random.seed(3)
        for layer in range(1, len(self.dimension_layers)):
            #parameters['W' + str(l)] = np.random.normal(0,1,(self.dimension_layers[l], self.dimension_layers[l - 1])) * 0.01
            parameters['W' + str(layer)] = np.random.randn(self.dimension_layers[layer], self.dimension_layers[layer - 1]) *0.1
            parameters['b' + str(layer)] = np.zeros((self.dimension_layers[layer], 1))

        return parameters





import setuptools


setuptools.setup(
    name='TensorFlow El 8lbah', # Replace with your own username
    version="0.0.1",
    author='MinaZ',
    author_email='mwagdy64@gmail.com',
    description='DL FrameWork to train and test on mnist and cifar10',
    long_description_content_type="text/markdown",
    url='https://github.com/Mina93178/Deep_learning_frame_work',
    packages=setuptools.find_packages(),
    install_requires=[
      'numpy', 'texttable', 'matplotlib',
    ],
  ## dependancy_links=['https://pypi.org/project/matplotlib/', 'https://pypi.org/project/numpy/', 'https://pypi.org/project/texttable/'],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        "Operating System :: OS Independent",
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',


    ],
      keywords='Deep_learning_frame_work ',
      python_requires='>=3.6',

)


import matplotlib.pyplot as plt
import matplotlib.animation as animation
import layers  , forward_model , Losses , backward_model
from dataset import dataset
from Metric import evaluation_metrics
from matplotlib import style
import shelve
import matplotlib; matplotlib.use("TkAgg")
style.use('fivethirtyeight')
fig = plt.figure()
ax1 = fig.add_subplot(1, 1, 1)
class training_model:
    '''
    " The Training class using batch gradient descent "
    '''
    def __init__(self,X, Y,layers_dimenstions,activation_functions, alpha=0.5 , no_of_iterations=60, print_cost = True, lambd = 0  ):
        '''
        :param X: The input of the first layer
        :param Y: the True labels of the training examples
        :param layers_dimenstions : the number of hidden units in each hidden layer
        :param activation_functions : the type of activation function in each hidden layer
        :param alpha: learning rate
        :param no_of_iterations: number of iterations
        :param print_cost: boolean variable , put it with ( True ) value if you want to print the cost every 10
                        iterations
        :param lambd: regularization parameter

        :return : The Trained Parameters for a certain model trained on a certain dataset
        '''
        self.input = X
        self.Y=Y
        self.learning_rate = alpha
        self.no_of_iterations = no_of_iterations
        self.print_cost=print_cost
        self.regularization_parameter=lambd
        self.dimension_layers = layers_dimenstions
        self.activation_functions = activation_functions

    def update_parameters(self,parameters, grads, learning_rate):
        '''
          "The function which is used to update the weights and biases of the model with Vanilla Gradient descent ."
        :param parameters: weights and biases of the past iteration ( before updating )
        :param grads: the Gradients of the weights and biases , the output of the backward propagation
        :param learning_rate: The learning rate
        :return: parameters : updated weights and biases after completing one iteration of the training
        '''
        L = len(parameters) // 2  # number of layers in the neural network
        for l in range(L):
            parameters["W" + str(l + 1)] = parameters["W" + str(l + 1)] - learning_rate * grads["dW" + str(l + 1)]
            parameters["b" + str(l + 1)] = parameters["b" + str(l + 1)] - learning_rate * grads["db" + str(l + 1)]

        return parameters

    def animate(i):
        '''
        "A function which is used to draw a live plotting of the Cost function during the training process "
        '''

        graph_data = open('costs.txt', 'r').read()
        lines = graph_data.split('\n')
        xs = []
        ys = []

        # plt.title("Learning rate =" + str(self.learning_rate))
        for line in lines:
            if len(line) > 1:
                x, y = line.split(',')
                xs.append(float(x))
                ys.append(float(y))
        ax1.clear()
        plt.ylabel('cost')
        plt.xlabel('iterations')
        ax1.plot(xs, ys)



    def save_model(self, filename):
        """Saves model by its name in the directory specified."""

        with shelve.open(filename) as db:
            db['self'] = self

    def train(self):

        '''
         " This function considered as the integration of all the past Modules together to start training any model
           the deep learning engineer will have the option to choose :
           1- the layers dimensions
           2- the activation function of each layer
           3- the loss type
           4- the number of iterations
          this function will plot live graph for the training cost and finally will print the the accuracy resulted
           from the test set training using the parameters resulted from the training set training .
        :return: The Trained parameters
        '''


        temp_layers= layers.layers(self.dimension_layers)
        # Initialize parameters dictionary.
        parameters = temp_layers.layers_init()
        #print (parameters)
        #print (parameters)
        #print("weights:")
        #print(parameters["W3"].shape)

        # Loop (gradient descent)
        cost_file = open("costs.txt", 'a+')
        cost_file.truncate(0)
        cost_file.close()
        for i in range(0,self.no_of_iterations):
            predictions, packet_of_packets = forward_model.forward_model().forward_model(self.input,parameters,self.activation_functions)
            #print(predictions.shape)
            #print (predictions)

            # Cost function
            if self.regularization_parameter == 0:
                cost =Losses.multiclass_loss(self.Y,predictions).cost()
                #print(cost)
            else:
                cost = Losses.regularization.compute_cost_with_regularization(predictions, self.Y, parameters, self.regularization_parameter,"multiclass")

            # Backward propagation.
            grads = backward_model.model_backward_general(predictions, self.Y, packet_of_packets, "multiclass",self.regularization_parameter,self.activation_functions).model_backward()
            # Update parameters.
            parameters = self.update_parameters(parameters, grads, self.learning_rate)
            # plot the cost
            #costs.append(cost)
            #self.visualize_cost(i,cost)
           # ani = animation.FuncAnimation(self.fig,self.draw,interval=1000)
            #plt.show()
            cost_file = open("costs.txt", 'a+')
            cost_file.write(f"{i},{cost} \n")
            cost_file.close()
            plt.ion()
            plt.ylabel('cost')
            plt.xlabel('iterations')
            plt.title("Learning rate =" + str(self.learning_rate))
            plt.show()
            plt.draw()
            plt.pause(1)
            '''
            plt.figure()
            plt.plot(costs)
            plt.ylabel('cost')
            plt.xlabel('iterations')
            plt.title("Learning rate =" + str(self.learning_rate))
            plt.show()
            '''
            # Print the loss every 10000 iterations
            if self.print_cost and i % 10 == 0:
                print("Cost after iteration {}: {}".format(i, cost))
        '''
        # plot the cost
        plt.figure()
        plt.plot(costs)
        plt.ylabel('cost')
        plt.xlabel('iterations')
        plt.title("Learning rate =" + str(self.learning_rate))
        plt.show()
        '''
        return parameters


def load_model(filename):
    """Loads a model saved in 'filename'."""

    with shelve.open(filename) as db:
        model = db['self']
    return model



if __name__ == "__main__":
    ani = animation.FuncAnimation(fig, training_model.animate, interval=1000)
    data_set = dataset('mnist', r'C:\Users\FacultyStudent\PycharmProjects\final_NN')
    train_X, train_Y, test_X, test_Y = data_set.get_dataset()
    temp = train_Y
    # print(train_Y.shape)
    train_Y = data_set.labels_to_onehot(train_Y)
    layers_dimensions = [train_X.T.shape[0], 128, train_Y.shape[0]]
    # print((train_X.shape))
    # print((train_Y.shape))
    # print(train_X.shape)
    # print(test_X.shape)
    # print(train_Y.shape)
    # print(test_Y.shape)
    # print("x_dimensions")
    # print(train_X.shape)
    # print("y_label:")
    # print(train_Y.shape)
    activation_functions = ["NONE", "sigmoid", "sigmoid"]
    parameters = training_model(train_X.T, train_Y, layers_dimensions, activation_functions).train()
    training_model(train_X.T, train_Y, layers_dimensions, activation_functions).save_model("test")
   # dectionary = training_model(train_X.T, train_Y, layers_dimensions, activation_functions).load_model("test.dir")
    dectionary = load_model("test.bak")
    print(dectionary)
    # parameters = training_model(X_train, Y_train).train()
    temp = evaluation_metrics(temp, train_X.T, parameters,activation_functions).Accuracy(train_X.shape[0])
    print(f"our accuracy : {temp} ")

    # print ("On the training set:")
    # predictions_train = predict(train_X.T, train_Y.T, parameters)
    # predictions_train = predict(train_X.T, train_Y, parameters)

    # print ("On the test set:")
    # predictions_test = predict(test_X.T, test_Y, parameters)
    # predictions_test = predict(X_test, Y_test, parameters)

    '''
    plt.title("Model without regularization")
    axes = plt.gca()
    axes.set_xlim([-0.75,0.40])
    axes.set_ylim([-0.75,0.65])
    plot_decision_boundary(lambda x: predict_dec(parameters, test_X.T), train_X, train_Y)
    '''



from init import initialization
class layers():
    '''
    This Class was just to keep the naming convention of the models , this will use the init function to get the weights
                & biases of the hidden layers .
    '''
    def __init__(self,layers_dimensions):
        self.layers_dimensions=layers_dimensions
        '''
                    " This function will initialize a random values for weights and zeros for biases using the
                                dimension_layers list "

                    :type layers_dimensions: list
                    :param layers_dimensions: list of the number of hidden units in each hidden layer . 

        '''
    def layers_init(self): # [5,4,6,7,1]
        init=initialization(self.layers_dimensions)
        return init.initialize_parameters()

'''
inputs = [1,6,4,7,1]
layerrrrrr=layers(inputs)
parameters = layerrrrrr.layers_init()
print(parameters)
'''


import numpy as np
import matplotlib.pyplot as plt
from forward_model import forward_model


import scipy.io


def sigmoid(x):
    """
    Compute the sigmoid of x
    Arguments:
    x -- A scalar or numpy array of any size.
    Return:
    s -- sigmoid(x)
    """
    s = 1 / (1 + np.exp(-x))
    return s


def relu(x):
    """
    Compute the relu of x
    Arguments:
    x -- A scalar or numpy array of any size.
    Return:
    s -- relu(x)
    """
    s = np.maximum(0, x)

    return s


def load_planar_dataset(seed):
    np.random.seed(seed)

    m = 400  # number of examples
    N = int(m / 2)  # number of points per class
    D = 2  # dimensionality
    X = np.zeros((m, D))  # data matrix where each row is a single example
    Y = np.zeros((m, 1), dtype='uint8')  # labels vector (0 for red, 1 for blue)
    a = 4  # maximum ray of the flower

    for j in range(2):
        ix = range(N * j, N * (j + 1))
        t = np.linspace(j * 3.12, (j + 1) * 3.12, N) + np.random.randn(N) * 0.2  # theta
        r = a * np.sin(4 * t) + np.random.randn(N) * 0.2  # radius
        X[ix] = np.c_[r * np.sin(t), r * np.cos(t)]
        Y[ix] = j

    X = X.T
    Y = Y.T

    return X, Y


def initialize_parameters(layer_dims):
    """
    Arguments:
    layer_dims -- python array (list) containing the dimensions of each layer in our network

    Returns:
    parameters -- python dictionary containing your parameters "W1", "b1", ..., "WL", "bL":
                    W1 -- weight matrix of shape (layer_dims[l], layer_dims[l-1])
                    b1 -- bias vector of shape (layer_dims[l], 1)
                    Wl -- weight matrix of shape (layer_dims[l-1], layer_dims[l])
                    bl -- bias vector of shape (1, layer_dims[l])

    Tips:
    - For example: the layer_dims for the "Planar Data classification model" would have been [2,2,1].
    This means W1's shape was (2,2), b1 was (1,2), W2 was (2,1) and b2 was (1,1). Now you have to generalize it!
    - In the for loop, use parameters['W' + str(l)] to access Wl, where l is the iterative integer.
    """

    np.random.seed(3)
    parameters = {}
    L = len(layer_dims)  # number of layers in the network

    for l in range(1, L):
        parameters['W' + str(l)] = np.random.randn(layer_dims[l], layer_dims[l - 1]) / np.sqrt(layer_dims[l - 1])
        parameters['b' + str(l)] = np.zeros((layer_dims[l], 1))

        assert (parameters['W' + str(l)].shape == (layer_dims[l], layer_dims[l - 1]))
        assert (parameters['b' + str(l)].shape == (layer_dims[l], 1))

    return parameters


def forward_propagation(X, parameters):
    """
    Implements the forward propagation (and computes the loss) presented in Figure 2.

    Arguments:
    X -- input dataset, of shape (input size, number of examples)
    parameters -- python dictionary containing your parameters "W1", "b1", "W2", "b2", "W3", "b3":
                    W1 -- weight matrix of shape ()
                    b1 -- bias vector of shape ()
                    W2 -- weight matrix of shape ()
                    b2 -- bias vector of shape ()
                    W3 -- weight matrix of shape ()
                    b3 -- bias vector of shape ()

    Returns:
    loss -- the loss function (vanilla logistic loss)
    """

    # retrieve parameters
    W1 = parameters["W1"]
    b1 = parameters["b1"]
    W2 = parameters["W2"]
    b2 = parameters["b2"]
    W3 = parameters["W3"]
    b3 = parameters["b3"]

    # LINEAR -> RELU -> LINEAR -> RELU -> LINEAR -> SIGMOID
    Z1 = np.dot(W1, X) + b1
    A1 = relu(Z1)
    Z2 = np.dot(W2, A1) + b2
    A2 = relu(Z2)
    Z3 = np.dot(W3, A2) + b3
    A3 = sigmoid(Z3)

    cache = (Z1, A1, W1, b1, Z2, A2, W2, b2, Z3, A3, W3, b3)

    return A3, cache


def backward_propagation(X, Y, cache):
    """
    Implement the backward propagation presented in figure 2.

    Arguments:
    X -- input dataset, of shape (input size, number of examples)
    Y -- true "label" vector (containing 0 if cat, 1 if non-cat)
    cache -- cache output from forward_propagation()

    Returns:
    gradients -- A dictionary with the gradients with respect to each parameter, activation and pre-activation variables
    """
    m = X.shape[1]
    (Z1, A1, W1, b1, Z2, A2, W2, b2, Z3, A3, W3, b3) = cache

   # dZ3 = np.multiply((A3 - Y),(sigmoid(Z3)*(1-sigmoid(Z3))))
    dZ3 = A3 - Y
    dW3 = 1. / m * np.dot(dZ3, A2.T)
    db3 = 1. / m * np.sum(dZ3, axis=1, keepdims=True)

    dA2 = np.dot(W3.T, dZ3)
    dZ2 = np.multiply(dA2, np.int64(A2 > 0))
    dW2 = 1. / m * np.dot(dZ2, A1.T)
    db2 = 1. / m * np.sum(dZ2, axis=1, keepdims=True)

    dA1 = np.dot(W2.T, dZ2)
    dZ1 = np.multiply(dA1, np.int64(A1 > 0))
    dW1 = 1. / m * np.dot(dZ1, X.T)
    db1 = 1. / m * np.sum(dZ1, axis=1, keepdims=True)

    gradients = {"dZ3": dZ3, "dW3": dW3, "db3": db3,
                 "dA2": dA2, "dZ2": dZ2, "dW2": dW2, "db2": db2,
                 "dA1": dA1, "dZ1": dZ1, "dW1": dW1, "db1": db1}

    return gradients


def update_parameters(parameters, grads, learning_rate):
    """
    Update parameters using gradient descent

    Arguments:
    parameters -- python dictionary containing your parameters:
                    parameters['W' + str(i)] = Wi
                    parameters['b' + str(i)] = bi
    grads -- python dictionary containing your gradients for each parameters:
                    grads['dW' + str(i)] = dWi
                    grads['db' + str(i)] = dbi
    learning_rate -- the learning rate, scalar.

    Returns:
    parameters -- python dictionary containing your updated parameters
    """

    n = len(parameters) // 2  # number of layers in the neural networks

    # Update rule for each parameter
    for k in range(n):
        parameters["W" + str(k + 1)] = parameters["W" + str(k + 1)] - learning_rate * grads["dW" + str(k + 1)]
        parameters["b" + str(k + 1)] = parameters["b" + str(k + 1)] - learning_rate * grads["db" + str(k + 1)]

    return parameters


def predict(X, y, parameters):

    """
    This function is used to predict the results of a  n-layer neural network.

    Arguments:
    X -- data set of examples you would like to label
    parameters -- parameters of the trained model

    Returns:
    p -- predictions for the given dataset X
    """

    m = X.shape[1]
    p = np.zeros((1, m), dtype=np.int)

    # Forward propagation
    a3, caches = forward_model().forward_model(X, parameters)

    # convert probas to 0/1 predictions
    for i in range(0, a3.shape[1]):
        if a3[0, i] > 0.5:
            p[0, i] = 1
        else:
            p[0, i] = 0

    # print results

    # print ("predictions: " + str(p[0,:]))
    # print ("true labels: " + str(y[0,:]))
    print("Accuracy: " + str(np.mean((p[0, :] == y[0, :]))))

    return p


def compute_cost(a3, Y):
    """
    Implement the cost function

    Arguments:
    a3 -- post-activation, output of forward propagation
    Y -- "true" labels vector, same shape as a3

    Returns:
    cost - value of the cost function
    """
    m = Y.shape[1]

    logprobs = np.multiply(-np.log(a3), Y) + np.multiply(-np.log(1 - a3), 1 - Y)
    cost = 1. / m * np.nansum(logprobs)

    return cost


def load_dataset():
    train_dataset = h5py.File('datasets/train_catvnoncat.h5', "r")
    train_set_x_orig = np.array(train_dataset["train_set_x"][:])  # your train set features
    train_set_y_orig = np.array(train_dataset["train_set_y"][:])  # your train set labels

    test_dataset = h5py.File('datasets/test_catvnoncat.h5', "r")
    test_set_x_orig = np.array(test_dataset["test_set_x"][:])  # your test set features
    test_set_y_orig = np.array(test_dataset["test_set_y"][:])  # your test set labels

    classes = np.array(test_dataset["list_classes"][:])  # the list of classes

    train_set_y = train_set_y_orig.reshape((1, train_set_y_orig.shape[0]))
    test_set_y = test_set_y_orig.reshape((1, test_set_y_orig.shape[0]))

    train_set_x_orig = train_set_x_orig.reshape(train_set_x_orig.shape[0], -1).T
    test_set_x_orig = test_set_x_orig.reshape(test_set_x_orig.shape[0], -1).T

    train_set_x = train_set_x_orig / 255
    test_set_x = test_set_x_orig / 255

    return train_set_x, train_set_y, test_set_x, test_set_y, classes


def predict_dec(parameters, X):
    """
    Used for plotting decision boundary.

    Arguments:
    parameters -- python dictionary containing your parameters
    X -- input data of size (m, K)

    Returns
    predictions -- vector of predictions of our model (red: 0 / blue: 1)
    """

    # Predict using forward propagation and a classification threshold of 0.5
    a3, cache = forward_model().forward_model(X, parameters)
    predictions = (a3 > 0.5)
    return predictions


def load_planar_dataset(randomness, seed):
    np.random.seed(seed)

    m = 50
    N = int(m / 2)  # number of points per class
    D = 2  # dimensionality
    X = np.zeros((m, D))  # data matrix where each row is a single example
    Y = np.zeros((m, 1), dtype='uint8')  # labels vector (0 for red, 1 for blue)
    a = 2  # maximum ray of the flower

    for j in range(2):

        ix = range(N * j, N * (j + 1))
        if j == 0:
            t = np.linspace(j, 4 * 3.1415 * (j + 1), N)  # + np.random.randn(N)*randomness # theta
            r = 0.3 * np.square(t) + np.random.randn(N) * randomness  # radius
        if j == 1:
            t = np.linspace(j, 2 * 3.1415 * (j + 1), N)  # + np.random.randn(N)*randomness # theta
            r = 0.2 * np.square(t) + np.random.randn(N) * randomness  # radius

        X[ix] = np.c_[r * np.cos(t), r * np.sin(t)]
        Y[ix] = j

    X = X.T
    Y = Y.T

    return X, Y


def plot_decision_boundary(model, X, y):
    # Set min and max values and give it some padding
    x_min, x_max = X[0, :].min() - 1, X[0, :].max() + 1
    y_min, y_max = X[1, :].min() - 1, X[1, :].max() + 1
    h = 0.01
    # Generate a grid of points with distance h between them
    xx, yy = np.meshgrid(np.arange(x_min, x_max, h), np.arange(y_min, y_max, h))
    # Predict the function value for the whole grid
    Z = model(np.c_[xx.ravel(), yy.ravel()])
    Z = Z.reshape(xx.shape)
    # Plot the contour and training examples
    plt.contourf(xx, yy, Z, cmap=plt.cm.Spectral)
    plt.ylabel('x2')
    plt.xlabel('x1')
    plt.scatter(X[0, :], X[1, :], c=y, cmap=plt.cm.Spectral)
    plt.show()


def load_2D_dataset():
    data = scipy.io.loadmat('datasets/data.mat')
    train_X = data['X'].T
    train_Y = data['y'].T
    test_X = data['Xval'].T
    test_Y = data['yval'].T

    plt.scatter(train_X[0, :], train_X[1, :], c=train_Y, s=40, cmap=plt.cm.Spectral);

    return train_X, train_Y, test_X, test_Y


import numpy as np
from layers import layers
import activations
class forward_prop:
    '''
    " This Class is Concerned with calculating the two parts of the forward propagation :
      identity part : calculating the product of the forward propagation before entering the activation function
      activation part : takes the output of the identity part and outputs the prediction based on the kind of the
                        activation function
    '''
    def identity_forward(self,X,W,b):
        '''
        :param X: the input of the current layer
        :param W: the weights of the current layer
        :param b: biases of the current layer
        :return: - The product of ( Weights * input ) + biases ,
                 - a Tuple which contains the Values of input , weights and biases of the current layer as to be used
                    in the backward propagation
        '''
        Z = np.dot(W, X) + b
       # print(Z)
        forward_packet = (X, W, b)
        return Z , forward_packet

    def activation_forward(self,input,W,b,activation_type):
        '''
        :param input: the input of the current layer
        :param W: the weights of the current layer
        :param b: biases of the current layer
        :param activation_type: Type of activation function used in the forward propagation
        :return: - A --> the output of the activation function
                 - packet_of_packets --> Tuple of 2 elements which will be used in backward propagation :
                     1- linear packer : contains ( input , weights , biases ) of the current layer
                     2- activation packet : contains ( Z ) which is the input to the activation function
        '''
        if activation_type == "sigmoid":
            Z, linear_packet = self.identity_forward(input, W, b) ## Z = input * w + b
            temp=activations.Sigmoid()
            A, activation_packet = temp.forward(Z) ## A = sig(z)

        elif activation_type == "relu":
            Z, linear_packet = self.identity_forward(input, W, b)
            temp = activations.relu()
            A, activation_packet = temp.forward(Z)

        elif activation_type == "leaky_relu":
            Z, linear_packet = self.identity_forward(input, W, b)
            temp = activations.leaky_relu()
            A, activation_packet = temp.forward(Z)
        elif activation_type == "tanh":
            Z, linear_packet = self.identity_forward(input, W, b)
            temp = activations.tanh()
            A, activation_packet = temp.forward(Z)
        elif activation_type == "softmax":
            Z, linear_packet = self.identity_forward(input, W, b)
            #temp =
            A, activation_packet = activations.Softmax().forward(Z)
        elif activation_type == "linear":
            Z, linear_packet = self.identity_forward(input, W, b)
            # temp =
            A, activation_packet = Z,Z

        else:
            raise ValueError("ERROR : Activation Function is Not Determined")

        packet_of_packets = linear_packet, activation_packet
        return A, packet_of_packets
    '''
    def activation_forward_with_droupout(self, input, W, b, droupouRatio,activation_type):
        if activation_type == "sigmoid":
            Z, linear_packet = self.identity_forward(input, W, b)  ## Z = input * w + b
            temp = activations.Sigmoid()
            A, activation_packet = temp.forward(Z)  ## A = sig(z)
            droup = np.random.randn(A.shape[0], A.shape[1])
            droup = (droup < droupouRatio)
            A = A * droup
            A = A / droupouRatio
            activation_packet = Z, droup

        elif activation_type == "relu":
            Z, linear_packet = self.identity_forward(input, W, b)
            temp = activations.relu()
            A, activation_packet = temp.forward(Z)
            droup=np.random.randn(A.shape[0],A.shape[1])
            droup=(droup<droupouRatio)
            A=A*droup
            A=A/droupouRatio
            activation_packet = Z,droup

        elif activation_type == "leaky_relu":
            Z, linear_packet = self.identity_forward(input, W, b)
            temp = activations.leaky_relu()
            A, activation_packet = temp.forward(Z)
            droup = np.random.randn(A.shape[0], A.shape[1])
            droup = (droup < droupouRatio)
            A = A * droup
            A = A / droupouRatio
            activation_packet = Z, droup

        elif activation_type == "tanh":
            Z, linear_packet = self.identity_forward(input, W, b)
            temp = activations.tanh()
            A, activation_packet = temp.forward(Z)
            droup = np.random.randn(A.shape[0], A.shape[1])
            droup = (droup < droupouRatio)
            A = A * droup
            A = A / droupouRatio
            activation_packet = Z, droup

        elif activation_type == "softmax":
            droupouRatio=1
            Z, linear_packet = self.identity_forward(input, W, b)
            # temp =
            A, activation_packet = activations.Softmax().forward(Z)
            droup = np.random.randn(A.shape[0], A.shape[1])
            droup = (droup < droupouRatio)
            A = A * droup
            A = A / droupouRatio
            activation_packet = Z, droup

        else:
            raise ValueError("pierre Was here")

        packet_of_packets = linear_packet, activation_packet
        return A, packet_of_packets

    '''
'''
test_cases=forward_prop()
np.random.seed(1)
A = np.random.randn(3, 2)
W = np.random.randn(1, 3)
b = np.random.randn(1, 1)
Z , forward_packet = test_cases.identity_forward(A,W,b)
print(forward_packet)
'''